#include"cola.h"

// 1) cOLA core:

SE *fquote(SE *arg,SE *env){return head(arg);}
SE *feval(SE *arg,SE *env){return eval(eval(head(arg),env),env);}
SE *fcond(SE *arg,SE *env)
{
  if(arg==NIL) return NIL;
  if(eval(head(head(arg)),env)!=NIL) return eval(head(tail(head(arg))),env);
  return fcond(tail(arg),env);
}

SE *crenv(SE *env,SE *bound,SE *expr) //////////////////////////
{
  SE *v;
  if(expr==NIL) return NIL;
  switch(type(expr))
    {
    case NUM:
    case STR:
    case FUN:
    case LAMBDA: return NIL;
    case SYM:
      if(ismem(expr,bound)) return NIL;
      if(assoc(expr,topenv)) return NIL;
      v=assoc(expr,env);
      if(v!=NIL) return new_cons(new_cons(expr,v),NIL);
       else return NIL;
    case CONS: return append(crenv(env,bound,head(expr)),
			     crenv(env,bound,tail(expr)));      
    }
  return NIL; /* notreached */
}

SE *flambda(SE *arg,SE *env)
{
  SE *al=head(arg),*bd=head(tail(arg));
  return new_lambda(al,bd,crenv(env,al,bd));
}

SE *flabel(SE *arg,SE *env)
{
  SE *s=head(arg),*e=eval(head(tail(arg)),env);
  ///////////////////
  SE *c;
  for(c=env;c!=NIL;c=tail(c))
    if(eq(head(head(c)),s))
    {
      destroy(tail(head(c)));
      tail(head(c))=clone(e);
      return e;
    }
  add2env(s,e);
  return assoc(s,topenv);
}

SE *feq(SE *arg,SE *env)
{
  if(eq(eval(head(arg),env),eval(head(tail(arg)),env)))
    return T;
  return NIL;
}

SE *fisnum(SE *arg,SE *env)
{
  if(isnum(eval(head(arg),env))) return T;
  return NIL;
}

SE *fisstr(SE *arg,SE *env)
{
  if(isstr(eval(head(arg),env))) return T;
  return NIL;
}

SE *fisatom(SE *arg,SE *env)
{
  if(eval(head(arg),env)==NIL) return T;
  if(isatom(eval(head(arg),env))) return T;
  return NIL;
}

SE *fcons(SE *arg,SE *env)
{
  return new_cons(eval(head(arg),env),eval(head(tail(arg)),env));
}

SE *fhead(SE *arg,SE *env)
{
  return head(eval(head(arg),env));
}

SE *ftail(SE *arg,SE *env)
{
  return tail(eval(head(arg),env));
}


// 2) arithmetics:

SE *fadd(SE *arg,SE *env)
{
  SE *a=eval(head(arg),env),
     *b=eval(head(tail(arg)),env);
  return new_num(numv(a)+numv(b));
}

SE *fsub(SE *arg,SE *env)
{
  SE *a=eval(head(arg),env),
     *b=eval(head(tail(arg)),env);
  return new_num(numv(a)-numv(b));
}

SE *fmul(SE *arg,SE *env)
{
  SE *a=eval(head(arg),env),
     *b=eval(head(tail(arg)),env);
  return new_num(numv(a)*numv(b));
}

SE *fdiv(SE *arg,SE *env)
{
  SE *a=eval(head(arg),env),
     *b=eval(head(tail(arg)),env);
  return new_num(numv(a)/numv(b));
}

SE *f_mod(SE *arg,SE *env)
{
  SE *a=eval(head(arg),env),
     *b=eval(head(tail(arg)),env);
  return new_num(numv(a)%numv(b));
}

SE *flt(SE *arg,SE *env)
{
  SE *a=eval(head(arg),env),
     *b=eval(head(tail(arg)),env);
  if(numv(a)<numv(b)) return T;
  return NIL;
}

SE *fgt(SE *arg,SE *env)
{
  SE *a=eval(head(arg),env),
     *b=eval(head(tail(arg)),env);
  if(numv(a)>numv(b)) return T;
  return NIL;
}

SE *fabsol(SE *arg,SE *env)
{
  int a=numv(eval(head(arg),env));
  if(a<0) return new_num(-a);
  return new_num(a);
}

// 3) standard:

SE *fapp(SE *arg,SE *env)
{
  SE *a=eval(head(arg),env),
     *b=eval(head(tail(arg)),env);
  return append(a,b);
}

SE *flen(SE *arg,SE *env)
{
  return new_num(len_(eval(head(arg),env)));
}


SE *frev(SE *arg,SE *env)
{
  return reverse(eval(head(arg),env));
}

SE *fmap(SE *arg,SE *env)
{
  SE *a=eval(head(arg),env),
     *b=eval(head(tail(arg)),env);
  return map(a,b,env);
}

SE *fpair(SE *arg,SE *env)
{
   SE *a=eval(head(arg),env),
      *b=eval(head(tail(arg)),env);
   return pair(a,b);
}

SE *fassoc(SE *arg,SE *env)
{
  SE *a=eval(head(arg),env),
     *b=eval(head(tail(arg)),env);
  return assoc(a,b);
}

SE *fevlis(SE *arg,SE *env)
{
  SE *a=eval(head(arg),env);
  return evlis(a,env);
}

SE *fismem(SE *arg,SE *env)
{
  if(ismem(head(arg),head(tail(arg)))) return T;
  return NIL;
}

// 4) strings
SE *fstrstr(SE *arg,SE *env)
{
   SE *a=eval(head(arg),env),
      *b=eval(head(tail(arg)),env);
   if(strstr(strv(a),strv(b))!=NULL) return T;
   return NIL;
}

SE *fstrcmp(SE *arg,SE *env)
{
   SE *a=eval(head(arg),env),
      *b=eval(head(tail(arg)),env);
   return new_num(strcmp(strv(a),strv(b)));
}

SE *fstrcat(SE *arg,SE *env)
{
   SE *a=eval(head(arg),env),
      *b=eval(head(tail(arg)),env);
   char buf[256],*as=strv(a),*bs=strv(b);
   sprintf(buf,"%s%s",as,bs);
   return new_str(strdup(buf));
}

SE *fstrlen(SE *arg,SE *env)
{
  char *s=strv(eval(head(arg),env));
  return new_num((int)strlen(s));
}

SE *fleft(SE *arg,SE *env)
{
   SE *a=eval(head(arg),env),
      *b=eval(head(tail(arg)),env);
   int n=numv(b),i=0;
   char buf[256],*s=strv(a);
   while(n-- && i<255 && *s!='\0')
     {
       buf[i++]=*s;
       s++;
     }
   buf[i]='\0';
   return new_str(strdup(buf));
}

SE *fright(SE *arg,SE *env)
{
  return NIL; ///////////////
}


SE *fread_(SE *arg,SE *env)
{
  // feee:
  printf("- ");
  return read();
}

SE *fdisp(SE *arg,SE *env)
{
  SE *a=eval(head(arg),env);
  display(a);
  printf("\n"); // a fe...
  return a;
}

SE *fload(SE *arg,SE *env)
{
  SE *a=eval(head(arg),env);
  FILE *f=fopen(strv(a),"r");
  if(f) a=loadse(f);
  else err("load failed");
  fclose(f);
  return a;
}

SE *fload2(SE *arg,SE *env)
{
  SE *e;
  arg=head(arg);
  FILE *f=fopen(strv(arg),"r");
  if(f==NULL) err("load failed");
  else
    {
      e=loadse(f);
      do
      {
	// gc??????
	//	printf("\n%s>",strv(arg));
	//	display(e);
	//        printf("\n");
	//display(eval(e,topenv));
eval(e,topenv);
	e=loadse(f);
      }while(e!=NIL && !(type(e)==STR && strcmp(strv(e),"#eof")==0));
      fclose(f);
      printf("zaladowano <%s>.\n",strv(arg));
      return T;
    }
  return NIL;
}

SE *fsave(SE *arg,SE *env) // teraz niewazny, bo tamten load... i tego...
{
  SE *a=eval(head(arg),env),*b=eval(head(tail(arg)),env);
  FILE *f=fopen(strv(a),"w");
  if(f) savese(f,b);
  else err("save failed");
  fclose(f);
  return T;
}

SE *fparse(SE *arg,SE *env)
{
  SE *a=eval(head(arg),env);
  return parse(strv(a));
}

SE *ftostr(SE *arg,SE *env)
{
  SE *a=eval(head(arg),env);
  return new_str(se2str(a));
}

// !!!!!!!!!!!!!!!
SE *flambdaenv(SE *arg,SE *env)
{
  if(arg==NIL) err("lambdaenv - () arg");
  arg=eval(head(arg),env);
  if(arg==NIL) err("lambdaenv - () arg");
  if(type(arg)==LAMBDA) return bind(arg);
    else err("lambdadenv - arg.is not <lambda>");
  return NIL; /* notreached */
}

SE *flambdalabel(SE *arg,SE *env)
{
  SE *lam=eval(head(arg),env),
     *atr=head(tail(arg)),
     *val=eval(head(tail(tail(arg))),env),
     *c;
  for(c=bind(lam);c!=NIL;c=tail(c))
    if(eq(head(head(c)),atr))
    {
      tail(head(c))=val;
      c=clone(bind(lam));
      break;
    }
  if(c==NIL) c=clone(new_cons(new_cons(atr,val),bind(lam)));
  destroy(bind(lam));
  bind(lam)=c;
  return assoc(atr,bind(lam));
}

SE *flambdaget(SE *arg,SE *env)
{
  SE *lam=head(arg),*atr=head(tail(atr));
  return assoc(atr,bind(lam));
}

SE *fshowenv(SE *arg,SE *env)
{
  return topenv;
}
