#include"cola.h"

SE *new_num(int n)
{
  SE *e=new(SE);
  if(e)
    {
      type(e)=NUM;
      numv(e)=n;
      gc_push(e);
    }
   else err("new_num failed");
  return e;
}

SE *new_sym(char *s)
{
  SE *e=new(SE);
  if(e)
    {
      type(e)=SYM;
      symv(e)=s;
      gc_push(e);
    }
   else err("new_sym failed");
  return e;
}

SE *new_str(char *s)
{
  SE *e=new(SE);
  if(e)
    {
      type(e)=STR;
      strv(e)=s;
      gc_push(e);
    }
   else err("new_str failed");
  return e;
}

SE *new_fun(SE *(*fun)(SE *,SE *))
{
  SE *e=new(SE);
  if(e)
    {
      type(e)=FUN;
      funv(e)=fun;
      // functions doesn't go to garbage collector's stack
    }
   else err("new_fun failed");
  return e;
}

SE *new_cons(SE *h,SE *t)
{
  SE *e=new(SE);
  if(e)
    {
      type(e)=CONS;
      head(e)=h;
      tail(e)=t;
      gc_push(e);
    }
   else err("new_cons failed");
  return e;
}

SE *new_lambda(SE *al,SE *bd,SE *env)
{
  SE *e=new(SE);
  if(e)
    {
      type(e)=LAMBDA;
      body(e)=clone(bd);
      argl(e)=clone(al);
      bind(e)=clone(env);
      gc_push(e);
    }
   else err("new_lambda failed");
  return e;
}

int eq(SE *a,SE *b)
{
  if(a==b) return 1;
  if(a==NIL || b==NIL) return 0;
  if(type(a)!=type(b)) return 0;
  switch(type(a))
    {
    case NUM: if(numv(a)==numv(b)) return 1;
               else return 0;
    case SYM: if(strcmp(symv(a),symv(b))==0) return 1;
               else return 0;
    case STR: if(strcmp(strv(a),strv(b))==0) return 1;
               else return 0;
    default: break;
    }
  return 0;
}

SE *clone(SE *e)
{
  SE *c=new(SE);
  if(e && c)
    {
      switch((type(c)=type(e)))
	{
	case NUM: numv(c)=numv(e); break;
	case SYM: symv(c)=strdup(symv(e)); break;
	case STR: strv(c)=strdup(strv(e)); break;
	case FUN: funv(c)=funv(e); break; // why clone function?
	case LAMBDA: argl(c)=clone(argl(e));
	             body(c)=clone(body(e));
	             bind(c)=clone(bind(e));
                     break;
	case CONS: head(c)=clone(head(e));
              	   tail(c)=clone(tail(e));
                   break;
	}
      return c;
    }
  return NIL;
}

void destroy(SE *e)
{
  if(e)
    {
      switch(type(e))
	{
	case SYM: free(symv(e)); break;
	case STR: free(strv(e)); break;
	case LAMBDA: destroy(argl(e));
	             destroy(body(e));
 	             destroy(bind(e)); break;
	case CONS: destroy(head(e));
      	           destroy(tail(e)); break;
	default: break;
	}
      free(e);
    }
}

